/*
 * $Id: screen.c,v 1.3 2003/02/09 07:34:16 kenta Exp $
 *
 * Copyright 2002 Kenta Cho. All rights reserved.
 */

/**
 * screen handler.
 *
 * @version $Revision: 1.3 $
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "noiz2sa.h"
#include "screen.h"
#include "clrtbl.h"
#include "vector.h"
#include "degutil.h"
#include "letterrender.h"
#include "attractmanager.h"

#include "gp2x/gp2x.h"

// use ARM optimized memset
#include "gp2x/asmmem.h"

int brightness = DEFAULT_BRIGHTNESS;

LayerBit *smokeBuf[lyrSize];	// pointers to pbuf
LayerBit pbuf[lyrSize];		// copy of l2buf
LayerBit l1buf[lyrSize];
LayerBit l2buf[lyrSize];

#ifndef WIZ
// set by gp2x.c
LayerBit *buf = 0;
LayerBit *lpbuf = 0, *rpbuf = 0;
#else
LayerBit buf_[LAYER_WIDTH * LAYER_HEIGHT];
LayerBit lpbuf_[PANEL_WIDTH * PANEL_HEIGHT];
LayerBit rpbuf_[PANEL_WIDTH * PANEL_HEIGHT];
LayerBit *buf = buf_;
LayerBit *lpbuf = lpbuf_;
LayerBit *rpbuf = rpbuf_;
#endif

// Handle BMP images.
#define SPRITE_NUM 7

//static SDL_Surface *sprite[SPRITE_NUM];
static LayerBit sprite[SPRITE_NUM][20*20];

static char *spriteFile[SPRITE_NUM] = {
  "title_n.bmp", "title_o.bmp", "title_i.bmp", "title_z.bmp", "title_2.bmp",
  "title_s.bmp", "title_a.bmp",
};

extern unsigned long keys;

static int find_nearset_color(unsigned int src)
{
  int min = 256, ind = 0, i, d;
  // first 16 clolors in palette is grayscale, so search there
  for (i = 0; i < 16; i++) {
    d = absN((int)src - (int)(color_palette[i] & 0xff));
    if (d < min) { min = d; ind = i; }
  }

  return ind;
}

static void loadSprites() {
  int i, u, j;
  char buffer[1200];
  FILE *f;

  for ( i=0 ; i<SPRITE_NUM ; i++ ) {
    strcpy(buffer, "data/images/");
    strcat(buffer, spriteFile[i]);
    f = fopen(buffer, "rb");
    if ( f == NULL ) {
      fprintf(stderr, "Unable to load: %s\n", buffer);
      exit(1);
    }
    fseek(f, 54, SEEK_SET); // skip bmp header
    fread(buffer, 1, 1200, f);
    for (u = 0; u < 20; u++) {
      for (j = 0; j < 20; j++) {
        sprite[i][u*20 + j] = find_nearset_color((unsigned) buffer[((19-u)*20 + j)*3]);
      }
    }
    fclose(f);
  }
}

void drawSprite(int n, int x, int y) {
  LayerBit *dst = buf + y*160 + x;
  LayerBit *src = sprite[n];
  int i;

  for (i = 20; i; i--, src += 20, dst += 160)
    memcpy(dst, src, 20);
}

// Initialize palletes.
static void initPalette() {
  int i, r, g, b;
  for ( i=0 ; i<256 ; i++ ) {
    b = color_palette[i];
    r = (b >> 16) & 0xff;
    g = (b >>  8) & 0xff;
    b =  b & 0xff;
    r = r*brightness/256;
    g = g*brightness/256;
    b = b*brightness/256;
    color_palette[i] = (r << 16) | (g << 8) | b;
  }
  gp2x_video_setpalette(color_palette, 256);
}


static void makeSmokeBuf() {
  static int zero = 0;
  int x, y, mx, my;

  for ( y=0 ; y<LAYER_HEIGHT ; y++ ) {
    for ( x=0 ; x<LAYER_WIDTH ; x++ ) {
      mx = x + sctbl[(x*8)&(DIV-1)]/128;
      my = y + sctbl[(y*8)&(DIV-1)]/128;
      if ( mx < 0 || mx >= LAYER_WIDTH || my < 0 || my >= LAYER_HEIGHT ) {
	smokeBuf[x+y*pitch] = (LayerBit *)&zero;
      } else {
	smokeBuf[x+y*pitch] = &(pbuf[mx+my*pitch]);
      }
    }
  }
}

void init_screen() {
  initPalette();
  makeSmokeBuf();
  clearLPanel();
  clearRPanel();

  loadSprites();
}

#ifndef USE_ASM
void blendScreen() {
  int i;
  for ( i = lyrSize-1 ; i >= 0 ; i-- ) {
    buf[i] = colorAlp[l1buf[i]][l2buf[i]];
  }
}
#endif

void prepareScreen() {
#ifndef USE_BLITTER
  LayerBit *src, *dst;
  int i;
#endif

  if ( status == TITLE ) {
    drawTitle();
  }

#ifdef USE_BLITTER
  // haven't noticed any problems without this, but just in case
  mmu_flushcache(gp2x_screen, (char *)gp2x_screen + 320*240);

  noiz2sa_blit();
#else
  // lpanel
  src = lpbuf;
  dst = gp2x_screen;
  for (i = PANEL_HEIGHT; i; i--, src+=80, dst+=320)
    memcpy(dst, src, 80);

  // buf
  src = buf;
  dst = (LayerBit *)gp2x_screen + 80;
  for (i = PANEL_HEIGHT; i; i--, src+=160, dst+=320)
    memcpy(dst, src, 160);

  // rpanel
  src = rpbuf;
  dst = (LayerBit *)gp2x_screen + 240;
  for (i = PANEL_HEIGHT; i; i--, src+=80, dst+=320)
    memcpy(dst, src, 80);

  mmu_flushcache(gp2x_screen, (char *)gp2x_screen + 320*240);
#endif
}

void clearLPanel() {
  // tried blitter here, no noticeable improvement
  memset(lpbuf, 0, 80*240);
}

void clearRPanel() {
  memset(rpbuf, 0, 80*240);
}

void clearFpsArea() {
  int i, *p;
  LayerBit *dst = lpbuf + 80*208;

  for (i = 32; i; i--, dst += 80) {
    p = (int *)dst; *p++ = 0; *p++ = 0; *p++ = 0; *p++ = 0; *p++ = 0; *p++ = 0; *p++ = 0; *p++ = 0;
  }
}

#ifndef USE_ASM
void smokeScreen() {
  int i;
  memcpy(pbuf, l2buf, lyrSize);
  for ( i = lyrSize-1 ; i >= 0 ; i-- ) {
    l1buf[i] = colorDfs[l1buf[i]];
    l2buf[i] = colorDfs[*(smokeBuf[i])];
  }
}
#endif

void drawBox(int x, int y, int width, int height,
	     LayerBit color1, LayerBit color2, LayerBit *buf) {
  int i;
  int ptr;

  x -= width>>1; y -= height>>1;
  if ( x < 0 ) {
    width += x; x = 0;
  }
  if ( x+width >= LAYER_WIDTH ) {
    width = LAYER_WIDTH-x;
  }
  if ( width <= 1 ) return;
  if ( y < 0 ) {
    height += y; y = 0;
  }
  if ( y+height > LAYER_HEIGHT ) {
    height = LAYER_HEIGHT-y;
  }
  if ( height <= 1 ) return;

  ptr = x + y*LAYER_WIDTH;
  memset(&(buf[ptr]), color2, width);
  y++;
  for ( i=0 ; i<height-2 ; i++, y++ ) {
    ptr = x + y*LAYER_WIDTH;
    buf[ptr] = color2; ptr++;
    memset(&(buf[ptr]), color1, width-2);
    ptr += width-2;
    buf[ptr] = color2;
  }
  ptr = x + y*LAYER_WIDTH;
  memset(&(buf[ptr]), color2, width);
}

void drawBoxPanel(int x, int y, int width, int height,
		  LayerBit color1, LayerBit color2, LayerBit *buf) {
  int i;
  int ptr;

  x -= width>>1; y -= height>>1;
  if ( x < 0 ) {
    width += x; x = 0;
  }
  if ( x+width >= PANEL_WIDTH ) {
    width = PANEL_WIDTH-x;
  }
  if ( width <= 1 ) return;
  if ( y < 0 ) {
    height += y; y = 0;
  }
  if ( y+height > PANEL_HEIGHT ) {
    height = PANEL_HEIGHT-y;
  }
  if ( height <= 1 ) return;

  ptr = x + y*PANEL_WIDTH;
  memset(&(buf[ptr]), color2, width);
  y++;
  for ( i=0 ; i<height-2 ; i++, y++ ) {
    ptr = x + y*PANEL_WIDTH;
    buf[ptr] = color2; ptr++;
    memset(&(buf[ptr]), color1, width-2);
    ptr += width-2;
    buf[ptr] = color2;
  }
  ptr = x + y*PANEL_WIDTH;
  memset(&(buf[ptr]), color2, width);
}

// Draw the numbers.
int drawNum(int n, int x ,int y, int s, int c1, int c2) {
  for ( ; ; ) {
    drawLetter(n%10, x, y, s, 1, c1, c2, lpbuf);
    // y += (s*1.8f) / (float)SCREEN_DIVISOR;
    y += s; // (s*480) >> 9;
    n /= 10;
    if ( n <= 0 ) break;
  }
  return y;
}

int drawNumRight(int n, int x ,int y, int s, int c1, int c2) {
  int d, nd, drawn = 0;
  for ( d = 100000000 ; d > 0 ; d /= 10 ) {
    nd = (int)(n/d);
    if ( nd > 0 || drawn ) {
      n -= d*nd;
      drawLetter(nd%10, x, y, s, 3, c1, c2, rpbuf);
      y += (s*1.7f) / (float)SCREEN_DIVISOR;
      drawn = 1;
    }
  }
  if ( !drawn ) {
    drawLetter(0, x, y, s, 3, c1, c2, rpbuf);
    y += (s*1.7f) / (float)SCREEN_DIVISOR;
  }
  return y;
}

int drawNumCenter(int n, int x ,int y, int s, int c1, int c2) {
  for ( ; ; ) {
    drawLetterBuf(n%10, x, y, s, 2, c1, c2, buf, 0);
    x -= (s*1.7f) / (float)SCREEN_DIVISOR;
    n /= 10;
    if ( n <= 0 ) break;
  }
  return y;
}

#ifdef DREAMCAST
#define JOYSTICK_AXIS 16
#else
#define JOYSTICK_AXIS 16384
#endif

int getPadState() {
  int pad = 0;

  if (keys & GP2X_UP)    pad |= PAD_UP;
  if (keys & GP2X_DOWN)  pad |= PAD_DOWN;
  if (keys & GP2X_LEFT)  pad |= PAD_LEFT;
  if (keys & GP2X_RIGHT) pad |= PAD_RIGHT;

  return pad;
}

int buttonReversed = 0;

int getButtonState() {
  int btn = 0;

  if (keys & (GP2X_X|GP2X_Y)) {
    if ( !buttonReversed ) {
      btn |= PAD_BUTTON2;
    } else {
      btn |= PAD_BUTTON1;
    }
  }

  if (keys & (GP2X_A|GP2X_B)) {
    if ( !buttonReversed ) {
      btn |= PAD_BUTTON1;
    } else {
      btn |= PAD_BUTTON2;
    }
  }

  return btn;
}
